!*****************************************************************************!
!* 
!*      $RCSfile: Flux_roe.f90,v $
!*
!*      $Revision: 1.4 $
!*
!*      $Date: 2004/01/29 09:58:07 $
!*
!*      $Author: iagmidu $
!*
!*****************************************************************************!

MODULE flux_roe_mod
  !--------------------------------------------------------------------------
  IMPLICIT NONE
  PRIVATE
  !----------------------------------------------------------------------------
  INTERFACE flux_roe
     MODULE PROCEDURE flux_roe
  END INTERFACE
  !----------------------------------------------------------------------------

  PUBLIC  :: flux_roe

CONTAINS

  SUBROUTINE flux_roe( gamma,                     &
       rho_l, rho_r, vn_l , vn_r, &
       vt_l , vt_r , p_l  , p_r , &
       flux_side                   )
    !--------------------------------------------------------------------------
    IMPLICIT NONE
    !--------------------------------------------------------------------------
    !argument list declaration
    REAL         :: gamma
    REAL         :: rho_l, rho_r
    REAL         :: vn_l , vn_r
    REAL         :: vt_l , vt_r
    REAL         :: p_l  , p_r
    REAL         :: flux_side(4)
    !local variable declaration
    REAL         :: e_r,e_l,H_r, H_l
    REAL         :: sqrt_rho_r, sqrt_rho_l
    REAL         :: vn_quer,vt_quer,H_quer,c_quer
    REAL         :: a1,a2,a3,a4
    REAL         :: r1(4),r2(4),r3(4),r4(4)
    REAL         :: delta_rho,delta_m1,delta_m2,delta_e,delta_eq
    REAL         :: gamma1,gamma2,gamma3,gamma4
    REAL         :: f1_r(4),f1_l(4)
    !--------------------------------------------------------------------------
    INTENT(IN)   :: gamma,                     &
         rho_l, rho_r, vn_l , vn_r, &
         vt_l , vt_r , p_l  , p_r 
    INTENT(OUT)  :: flux_side    
    !--------------------------------------------------------------------------

    ! Calculate left/right energy and enthalpy

    e_r=1./(gamma-1.)*p_r+0.5*rho_r*(vn_r*vn_r+vt_r*vt_r)
    e_l=1./(gamma-1.)*p_l+0.5*rho_l*(vn_l*vn_l+vt_l*vt_l)

    H_r=(e_r + p_r)/rho_r
    H_l=(e_l + p_l)/rho_l


    ! Calculate sqrt(rho)

    sqrt_rho_r=SQRT(rho_r)
    sqrt_rho_l=SQRT(rho_l)

    ! Calculate Roe mean values

    vn_quer=(sqrt_rho_r*vn_r+sqrt_rho_l*vn_l)/(sqrt_rho_r+sqrt_rho_l)
    vt_quer=(sqrt_rho_r*vt_r+sqrt_rho_l*vt_l)/(sqrt_rho_r+sqrt_rho_l)
    H_quer=(sqrt_rho_r* H_r+sqrt_rho_l* H_l)/(sqrt_rho_r+sqrt_rho_l)

    c_quer=SQRT((gamma-1.)*(H_quer-0.5*(vn_quer*vn_quer + vt_quer*vt_quer)))

    ! Calculate mean eigenvalues

    a1=vn_quer-c_quer
    a2=vn_quer
    a3=vn_quer
    a4=vn_quer+c_quer

    ! Calculate mean eigenvectors

    r1(1)=1
    r1(2)=vn_quer-c_quer
    r1(3)=vt_quer
    r1(4)=H_quer-vn_quer*c_quer

    r2(1)=1
    r2(2)=vn_quer
    r2(3)=vt_quer
    r2(4)=0.5*(vn_quer*vn_quer + vt_quer*vt_quer)

    r3(1)=0
    r3(2)=0
    r3(3)=1
    r3(4)=vt_quer

    r4(1)=1
    r4(2)=vn_quer+c_quer
    r4(3)=vt_quer
    r4(4)=H_quer+vn_quer*c_quer

    ! Calculate differences

    delta_rho = rho_r      - rho_l
    delta_m1  = rho_r*vn_r - rho_l*vn_l
    delta_m2  = rho_r*vt_r - rho_l*vt_l
    delta_e   = e_r        - e_l
    delta_eq  = delta_e-(delta_m2-vt_quer*delta_rho)*vt_quer

    ! Calculate factors gamma1,...,gamma4

    gamma2 = -(gamma-1.)/(c_quer*c_quer)* &
         ( delta_rho*(vn_quer*vn_quer-H_quer) &
         + delta_eq &
         - delta_m1*vn_quer )
    gamma1 = -0.5/c_quer*( delta_m1 - delta_rho*(vn_quer+c_quer) )-0.5*gamma2
    gamma4 = delta_rho - gamma1 - gamma2 
    gamma3 = delta_m2 - vt_quer*delta_rho

    ! Calculalte the physical fluxes

    f1_r(1) = rho_r*vn_r
    f1_r(2) = rho_r*vn_r*vn_r + p_r
    f1_r(3) = rho_r*vn_r*vt_r
    f1_r(4) = vn_r*(e_r + p_r)

    f1_l(1) = rho_l*vn_l
    f1_l(2) = rho_l*vn_l*vn_l + p_l
    f1_l(3) = rho_l*vn_l*vt_l
    f1_l(4) = vn_l*(e_l + p_l)

    ! Calculate Roe fluxes

    flux_side(:) = 0.5*( f1_r(:)+f1_l(:) ) & 
         - 0.5*gamma1*ABS(a1)*r1(:) &
         - 0.5*gamma2*ABS(a2)*r2(:) &
         - 0.5*gamma3*ABS(a3)*r3(:) &
         - 0.5*gamma4*ABS(a4)*r4(:) 

    RETURN
  END SUBROUTINE flux_roe

END MODULE flux_roe_mod
