!*****************************************************************************!
!* 
!*      $RCSfile: SetBoundaryCondition.F90,v $
!*
!*      $Revision: 1.6 $
!*
!*      $Date: 2004/02/09 18:32:56 $
!*
!*      $Author: roller $
!*
!*****************************************************************************!

MODULE SetBoundaryCondition_mod
  !---------------------------------------------------------------------------!
  IMPLICIT NONE
  PRIVATE
  !---------------------------------------------------------------------------!
  INTERFACE IniBoundaryCondition
     MODULE PROCEDURE IniBoundaryCondition
  END INTERFACE
  INTERFACE SetBoundaryCondition
     MODULE PROCEDURE SetBoundaryCondition
  END INTERFACE
  INTERFACE CloseBoundaryCondition
     MODULE PROCEDURE CloseBoundaryCondition
  END INTERFACE

  REAL,POINTER            :: sendleft(:,:), sendright(:,:)  ! Randwerte-Vektor
  REAL,POINTER            :: sendup(:,:), senddown(:,:)     ! Randwerte-Vektor
  REAL,POINTER            :: getleft(:,:), getright(:,:)    ! Randwerte-Vektor
  REAL,POINTER            :: getupper(:,:), getlower(:,:)   ! Randwerte-Vektor
  INTEGER                 :: allocStat

  !---------------------------------------------------------------------------!
  PUBLIC  :: SetBoundaryCondition,IniBoundaryCondition,CloseBoundaryCondition
  !---------------------------------------------------------------------------!

CONTAINS

  SUBROUTINE SetBoundaryCondition(pvar,Const,Mesh,MPI)
    USE Types
    !---------------------------------------------------------------------------!
    IMPLICIT NONE
    !---------------------------------------------------------------------------!
#ifdef parallel
    include 'mpif.h'
#endif
    !---------------------------------------------------------------------------!
    ! Argument list declaration
    TYPE(tConstants)        :: Const
    TYPE(tMesh)             :: Mesh
    TYPE(tMPI)              :: MPI
    REAL                    :: pvar(0:Mesh%IMAX+1,0:Mesh%JMAX+1,Const%nvar)
    ! Local variable declaration
#ifdef parallel
    integer                 :: left, right                    ! Nachbarprozesse
    integer                 :: rqleft,rqright,rqupper,rqlower ! Request-Handles
    integer                 :: status(MPI_STATUS_SIZE)
#endif
    !---------------------------------------------------------------------------!
    INTENT(IN)              :: Const, Mesh
    INTENT(INOUT)           :: pvar
    !---------------------------------------------------------------------------!

    ! Implemented Boundary Conditions:
    ! Const%BoundCond = 1: Extrapolation 1st order at all boundaries
    ! Const%BoundCond = 2: Extrapolation 1st order in x, periodic in y-direction
    ! Const%BoundCond = 3: Periodic in x-direction, extrapolation 1st order in y
    ! Const%BoundCond = 4: Periodic in x- and y-direction


    ! Set left boundary

#ifdef parallel
    call mpi_cart_shift(MPI%comm_cart,0,1,left,right,MPI%ierror)

    ! Umkopieren: sendright = Vektor der letzen inneren Spalte
    sendright(1:Mesh%JMAX,:) = pvar(Mesh%IMAX,1:Mesh%JMAX,:)

    ! Sende sendright an den rechten Nachbarn
    ! Bin ich am rechten Rechengebietsrand und RB=periodisch,
    ! dann ist mein rechter Nachbar der am linken Rechengebietsrand
    ! Bin ich am rechten Rechengebietsrand und RB=Extrapolation,
    ! dann sendet das ins Leere (nach MPI_PROC_NULL). Das strt nicht.

    call mpi_issend(sendright,Mesh%JMAX*Const%nvar,mpi_real,right,17, &
         MPI%comm_cart,rqright,MPI%ierror)

    ! Empfange vom linken Nachbarn
    ! Bin ich am linken Rechengebietsrand und RB=periodisch,
    ! dann ist das der Prozess am rechten Rechengebietsrand
    ! Bin ich am linken Rechengebietsrand und RB=Extrapolation,
    ! dann empfange ich aus dem Nichts (aus MPI_PROC_NULL). Strt auch nicht.
    
    call mpi_recv(getleft,Mesh%JMAX*Const%nvar,mpi_real,left,17, &
         MPI%comm_cart,status,MPI%ierror)
    call mpi_wait(rqright,status,MPI%ierror)

    ! Zurckspeichern:
    pvar(0,1:Mesh%JMAX,:) = getleft(1:Mesh%JMAX,:)
    
    ! Bin ich echter Rand und RB=Extrapolation, 
    ! dann mu ich jetzt noch das Richtige setzen:

    if (MPI%coords(1) == 0) then
       IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.2)) THEN
          pvar(0,:,:) = pvar(1,:,:)                       ! Extrapolation 1st order 
       endif
    endif

#else
       IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.2)) THEN
          pvar(0,:,:) = pvar(1,:,:)                       ! Extrapolation 1st order 
       ELSEIF ((Const%BoundCond.eq.3).or.(Const%BoundCond.eq.4)) THEN
          pvar(0,:,:) = pvar(MPI%IMAX,:,:)               ! Periodic BC in x-direction
       ELSE
          write(*,*) 'No other boundary condition known'
          STOP
       ENDIF
#endif

    ! Set right boundary

#ifdef parallel
    ! Umkopieren: sendleft = Vektor der ersten inneren Spalte
    sendleft(1:Mesh%JMAX,:) = pvar(1,1:Mesh%JMAX,:)

    ! Sende sendleft an den linken Nachbarn

    call mpi_issend(sendleft,Mesh%JMAX*Const%nvar,mpi_real,left,18, &
         MPI%comm_cart,rqleft,MPI%ierror)

    ! Empfange vom linken Nachbarn
    ! Bin ich am linken Rechengebietsrand und RB=periodisch,
    ! dann ist das der Prozess am rechten Rechengebietsrand
    ! Bin ich am linken Rechengebietsrand und RB=Extrapolation,
    ! dann empfange ich aus dem Nichts (aus MPI_PROC_NULL). Strt auch nicht.
    
    call mpi_recv(getright,Mesh%JMAX*Const%nvar,mpi_real,right,18, &
         MPI%comm_cart,status,MPI%ierror)
    call mpi_wait(rqleft,status,MPI%ierror)

    ! Zurckspeichern:
    pvar(Mesh%IMAX+1,1:Mesh%JMAX,:) = getright(1:Mesh%JMAX,:)
    
    ! Bin ich echter Rand und RB=Extrapolation, 
    ! dann mu ich jetzt noch das Richtige setzen:

    if (MPI%coords(1) == (MPI%dims(1)-1)) then
       IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.2)) THEN
          pvar(Mesh%IMAX+1,:,:) = pvar(Mesh%IMAX,:,:)     ! Extrapolation 1st order 
       endif
    endif

#else
    IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.2)) THEN
       pvar(Mesh%IMAX+1,:,:) = pvar(Mesh%IMAX,:,:)     ! Extrapolation 1st order 
    ELSEIF ((Const%BoundCond.eq.3).or.(Const%BoundCond.eq.4)) THEN
       pvar(Mesh%IMAX+1,:,:) = pvar(1,:,:)             ! Periodic BC in x-direction
    ELSE
       write(*,*) 'No other boundary condition known'
       STOP
    ENDIF
#endif

    ! Set lower boundary

#ifdef parallel
    call mpi_cart_shift(MPI%comm_cart,1,1,left,right,MPI%ierror)

    ! Umkopieren: sendup = Vektor der letzen inneren Zeile
    sendup(1:Mesh%IMAX,:) = pvar(1:Mesh%IMAX,Mesh%JMAX,:)

    ! Sende sendup an den oberen Nachbarn

    call mpi_issend(sendup,Mesh%IMAX*Const%nvar,mpi_real,right,19, &
         MPI%comm_cart,rqupper,MPI%ierror)

    ! Empfange vom unteren Nachbarn
    
    call mpi_recv(getlower,Mesh%IMAX*Const%nvar,mpi_real,left,19, &
         MPI%comm_cart,status,MPI%ierror)
    call mpi_wait(rqupper,status,MPI%ierror)

    ! Zurckspeichern:
    pvar(1:Mesh%IMAX,0,:) = getlower(1:Mesh%IMAX,:)
    
    ! Bin ich echter Rand und RB=Extrapolation, 
    ! dann mu ich jetzt noch das Richtige setzen:

    if (MPI%coords(2) == 0) then
       IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.3)) THEN
          pvar(:,0,:) = pvar(:,1,:)                       ! Extrapolation 1st order 
       endif
    endif

#else
    IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.3)) THEN
       pvar(:,0,:) = pvar(:,1,:)                       ! Extrapolation 1st order 
    ELSEIF ((Const%BoundCond.eq.2).or.(Const%BoundCond.eq.4)) THEN
       pvar(:,0,:) = pvar(:,MESH%JMAX,:)               ! Periodic BC in y-direction
    ELSE
       write(*,*) 'No other boundary condition known'
       STOP
    ENDIF
#endif

    ! Set upper boundary

#ifdef parallel
    ! Umkopieren: senddown = Vektor der ersten inneren Zeile
    senddown(1:Mesh%IMAX,:) = pvar(1:Mesh%IMAX,1,:)

    ! Sende sendup an den oberen Nachbarn

    call mpi_issend(senddown,Mesh%IMAX*Const%nvar,mpi_real,left,20, &
         MPI%comm_cart,rqlower,MPI%ierror)

    ! Empfange vom unteren Nachbarn
    
    call mpi_recv(getupper,Mesh%IMAX*Const%nvar,mpi_real,right,20, &
         MPI%comm_cart,status,MPI%ierror)
    call mpi_wait(rqlower,status,MPI%ierror)

    ! Zurckspeichern:
    pvar(1:Mesh%IMAX,Mesh%JMAX+1,:) = getupper(1:Mesh%IMAX,:)
    
    ! Bin ich echter Rand und RB=Extrapolation, 
    ! dann mu ich jetzt noch das Richtige setzen:

    if (MPI%coords(2) == (MPI%dims(2)-1)) then
       IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.3)) THEN
          pvar(:,Mesh%JMAX+1,:) = pvar(:,Mesh%JMAX,:)     ! Extrapolation 1st order 
       endif
    endif

#else
    IF ((Const%BoundCond.eq.1).or.(Const%BoundCond.eq.3)) THEN
       pvar(:,Mesh%JMAX+1,:) = pvar(:,Mesh%JMAX,:)     ! Extrapolation 1st order 
    ELSEIF ((Const%BoundCond.eq.2).or.(Const%BoundCond.eq.4)) THEN
       pvar(:,Mesh%JMAX+1,:) = pvar(:,1,:)             ! Periodic BC in y-direction
    ELSE
       write(*,*) 'No other boundary condition known'
       STOP
    ENDIF
#endif
  
  END SUBROUTINE SetBoundaryCondition


  SUBROUTINE IniBoundaryCondition(Const,Mesh)
    USE Types
    !---------------------------------------------------------------------------!
    IMPLICIT NONE
    !---------------------------------------------------------------------------!
    ! Argument list declaration
    TYPE(tConstants)        :: Const
    TYPE(tMesh)             :: Mesh
    !---------------------------------------------------------------------------!
    INTENT(IN)              :: Const, Mesh
    !---------------------------------------------------------------------------!

#ifdef parallel
    ALLOCATE(sendleft(1:Mesh%JMAX,Const%nvar),   &
             sendright(1:Mesh%JMAX,Const%nvar),  &
             sendup(1:Mesh%IMAX,Const%nvar),   &
             senddown(1:Mesh%IMAX,Const%nvar),  &
             getleft(1:Mesh%JMAX,Const%nvar),   &
             getright(1:Mesh%JMAX,Const%nvar),  &
             getupper(1:Mesh%IMAX,Const%nvar),   &
             getlower(1:Mesh%IMAX,Const%nvar),  &
             STAT = allocStat )
    IF(allocStat.NE.0) THEN
       PRINT *, 'ERROR IniBoundaryCondition: ', &
                'Could not allocate all variables. '
       STOP
    ENDIF

    getleft  = 0.
    getright = 0.
    getupper = 0.
    getlower = 0.

#endif

  END SUBROUTINE IniBoundaryCondition

  SUBROUTINE CloseBoundaryCondition()
    USE Types
    !---------------------------------------------------------------------------!
    IMPLICIT NONE
    !---------------------------------------------------------------------------!
    ! Argument list declaration
    ! Local variable declaration
    !---------------------------------------------------------------------------!

#ifdef parallel
    DEALLOCATE(sendleft, sendright, getleft, getright,  &
             STAT = allocStat )
    IF(allocStat.NE.0) THEN
       PRINT *, 'ERROR CloseBoundaryCondition: ', &
                'Could not deallocate all variables. '
       STOP
    ENDIF

#endif
  END SUBROUTINE CloseBoundaryCondition

END MODULE SetBoundaryCondition_mod
